import creole

from creole.html_emitter import HtmlEmitter

from .utils import binders_map, object_slug


class PinaxWikiHtmlEmitter(HtmlEmitter):

    def __init__(self, wiki, root, link_rules=None):
        self.wiki = wiki
        super(PinaxWikiHtmlEmitter, self).__init__(root, link_rules)

    def link_emit(self, node):
        target = node.content
        if node.children:
            inside = self.emit_children(node)
        else:
            inside = self.html_escape(target)
        m = self.link_rules.addr_re.match(target)
        if m:
            if m.group('extern_addr'):
                return u'<a href="%s">%s</a>' % (
                    self.attr_escape(target), inside)
            elif m.group('inter_wiki'):
                raise NotImplementedError
        slug = object_slug(self.wiki)
        page_url = binders_map()[slug].page_url(self.wiki, target)
        return u'<a href="{0}">{1}</a>'.format(self.attr_escape(page_url), inside)


def creole_parse(wiki, text):
    return PinaxWikiHtmlEmitter(wiki, creole.Parser(text).parse()).emit()


def creole_wikiword_parse(wiki, text):
    rules = creole.Rules(wiki_words=True)
    return PinaxWikiHtmlEmitter(wiki, creole.Parser(text, rules).parse()).emit()



from django import forms

from .models import Revision


class RevisionForm(forms.ModelForm):

    revision_pk = forms.IntegerField(required=False, widget=forms.HiddenInput())
    message = forms.CharField(required=False, help_text="Leave a helpful message about your change")

    def __init__(self, *args, **kwargs):
        self.revision = kwargs.pop("revision")
        super(RevisionForm, self).__init__(*args, **kwargs)
        if self.revision:
            self.fields["content"].initial = self.revision.content
            self.fields["revision_pk"].initial = self.revision.pk
        else:
            self.fields["content"].initial = "add content and create a new page"
            self.fields["message"].initial = "initial revision"

    def clean_content(self):
        if self.revision and self.cleaned_data["content"] == self.revision.content:
            raise forms.ValidationError("You made no stinking changes")
        return self.cleaned_data["content"]

    def clean(self):
        if self.revision and self.cleaned_data.get("revision_pk") != self.revision.pk:
            raise forms.ValidationError("Someone edited this before you")
        return self.cleaned_data

    class Meta:
        model = Revision
        fields = [
            "revision_pk",
            "content",
            "message"
        ]



from django import template

from django.contrib.contenttypes.models import ContentType

from ..models import Wiki


register = template.Library()


@register.assignment_tag
def wiki_for(obj):
    return Wiki.objects.get(
        object_id=obj.pk,
        content_type=ContentType.objects.get_for_model(obj)
    )



from django.db.models.signals import post_save

from django.contrib.contenttypes.models import ContentType

from .conf import settings
from .models import Wiki


def handle_object_save(sender, **kwargs):
    created = kwargs.pop("created")
    obj = kwargs.pop("instance")
    if created:
        Wiki.objects.create(
            content_type=ContentType.objects.get_for_model(obj),
            object_id=obj.pk
        )


for binder in settings.WIKI_BINDERS:
    if binder.bind_to_model:
        post_save.connect(handle_object_save, sender=binder.bind_to_model)



import os

from django.conf.urls import patterns, url

from .conf import settings


urlpatterns = patterns(
    "wiki.views",
    url(r"^file-download/(\d+)/([^/]+)$", "file_download", name="wiki_file_download"),
    url(r"^file-upload/$", "file_upload", name="wiki_file_upload")
)

for binder in settings.WIKI_BINDERS:
    urlpatterns += patterns(
        "wiki.views",
        url(os.path.join(binder.root, r"$"), "index", {"binder": binder}, name=binder.index_url_name),
        url(os.path.join(binder.root, r"(?P<slug>[^/]+)/$"), "page", {"binder": binder}, name=binder.page_url_name),
        url(os.path.join(binder.root, r"(?P<slug>[^/]+)/edit/$"), "edit", {"binder": binder}, name=binder.edit_url_name),
    )



from .conf import settings


def binders_map():
    return {
        x.bind_to_model_name: x
        for x in settings.WIKI_BINDERS
    }


def object_slug(wiki):
    if wiki:
        slug = wiki.content_type.model
    else:
        slug = "default"
    return slug



from django.core.urlresolvers import reverse

from django.contrib.contenttypes.models import ContentType


class Binder(object):

    bind_to_model = None

    @property
    def bind_to_model_name(self):
        return self.bind_to_model._meta.model_name

    @property
    def index_url_name(self):
        return "{0}_wiki_index".format(self.bind_to_model_name)

    @property
    def page_url_name(self):
        return "{0}_wiki_page".format(self.bind_to_model_name)

    @property
    def edit_url_name(self):
        return "{0}_wiki_edit".format(self.bind_to_model_name)

    def get_object(self, **kwargs):
        return self.bind_to_model._default_manager.get(**{
            self.slug_name: kwargs.get(self.slug_name)
        })

    def lookup(self, *args, **kwargs):
        from .models import Wiki
        obj = self.get_object(**kwargs)
        return Wiki.objects.get(
            content_type=ContentType.objects.get_for_model(obj),
            object_id=obj.pk
        )


class DefaultBinder(Binder):

    @property
    def bind_to_model_name(self):
        return "default"

    @property
    def root(self):
        return r""

    def lookup(self, *args, **kwargs):
        return None

    @property
    def index_url_name(self):
        return "wiki_index"

    @property
    def page_url_name(self):
        return "wiki_page"

    @property
    def edit_url_name(self):
        return "wiki_edit"

    def page_url(self, wiki, slug):
        return reverse("wiki_page", kwargs={"slug": slug})

    def page_edit_url(self, wiki, slug):
        return reverse("wiki_edit", kwargs={"slug": slug})



import json

from django.http import HttpResponse, Http404, HttpResponseForbidden
from django.shortcuts import redirect, render, get_object_or_404
from django.views import static
from django.views.decorators.http import require_POST

from account.decorators import login_required

from .conf import settings
from .forms import RevisionForm
from .hooks import hookset
from .models import Page, MediaFile


def index(request, binder, *args, **kwargs):
    wiki = binder.lookup(*args, **kwargs)
    return redirect(binder.page_url(wiki, "WikiIndex"))


def page(request, slug, binder, *args, **kwargs):
    wiki = binder.lookup(*args, **kwargs)
    try:
        if wiki:
            page = wiki.pages.get(slug=slug)
        else:
            page = Page.objects.get(slug=slug)
        if not hookset.can_view_page(page, request.user):
            raise Http404()
        rev = page.revisions.latest()
        return render(request, "wiki/page.html", {"revision": rev, "can_edit": hookset.can_edit_page(page, request.user)})
    except Page.DoesNotExist:
        return redirect(binder.edit_url(wiki, slug))


def edit(request, slug, binder, *args, **kwargs):
    wiki = binder.lookup(*args, **kwargs)
    try:
        if wiki:
            page = wiki.pages.get(slug=slug)
        else:
            page = Page.objects.get(slug=slug)
        rev = page.revisions.latest()
        if not hookset.can_edit_page(page, request.user):
            return HttpResponseForbidden()
    except Page.DoesNotExist:
        page = Page(wiki=wiki, slug=slug)
        rev = None
        if not hookset.can_edit_page(page, request.user):
            raise Http404()
    if request.method == "POST":
        form = RevisionForm(request.POST, revision=rev)
        if form.is_valid():
            if page.pk is None:
                page.save()
            revision = form.save(commit=False)
            revision.page = page
            revision.created_by = request.user
            revision.created_ip = request.META.get(settings.WIKI_IP_ADDRESS_META_FIELD, "REMOTE_ADDR")
            revision.parse()
            revision.save()
            return redirect(binder.page_url(wiki, slug))
    else:
        form = RevisionForm(revision=rev)

    return render(request, "wiki/edit.html", {
        "form": form,
        "page": page,
        "revision": rev,
        "can_delete": hookset.can_delete_page(page, request.user)
    })


def file_download(request, pk, filename):
    media_file = get_object_or_404(MediaFile, pk=pk, filename=filename)
    if getattr(settings, "DOCUMENTS_USE_X_ACCEL_REDIRECT", False):
        response = HttpResponse()
        response["X-Accel-Redirect"] = media_file.file.url
        # delete content-type to allow Gondor to determine the filetype and
        # we definitely don't want Django's crappy default :-)
        del response["content-type"]
    else:
        response = static.serve(request, media_file.file.name, document_root=settings.MEDIA_ROOT)
    return response


@require_POST
@login_required
def file_upload(request):
    uploads = []
    for f in request.FILES.getlist("files"):
        media_file = request.user.media_files.create(file=f, filename=f.name)
        uploads.append(media_file)
    return HttpResponse(json.dumps({
        "uploads": [
            {"filename": m.filename, "download_url": m.download_url()}
            for m in uploads
        ]
    }), content_type="application/json")



import os
import re
import uuid

from django.core.urlresolvers import reverse
from django.db import models
from django.utils import timezone

from django.contrib.auth.models import User
from django.contrib.contenttypes.generic import GenericForeignKey
from django.contrib.contenttypes.models import ContentType

from .conf import settings
from .utils import binders_map, object_slug


MEDIA_RE = re.compile(r"w/file-download/(\d+)/")


class Wiki(models.Model):
    content_type = models.ForeignKey(ContentType)
    object_id = models.IntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    def latest_edits(self):
        return Revision.objects.filter(page__wiki=self).order_by("-created_at")

    class Meta:
        unique_together = [("content_type", "object_id")]


class Page(models.Model):
    wiki = models.ForeignKey(Wiki, related_name="pages", null=True)
    slug = models.SlugField()

    def get_absolute_url(self):
        slug = object_slug(self.wiki)
        return binders_map()[slug].page_url(self.wiki, self.slug)

    def get_edit_url(self):
        slug = object_slug(self.wiki)
        return binders_map()[slug].edit_url(self.wiki, self.slug)

    class Meta:
        unique_together = [("wiki", "slug")]


class Revision(models.Model):
    page = models.ForeignKey(Page, related_name="revisions")
    content = models.TextField(help_text="Use markdown to mark up your text")
    content_html = models.TextField()
    message = models.TextField(blank=True, help_text="Leave a helpful message about your change")
    created_ip = models.IPAddressField()
    created_at = models.DateTimeField(default=timezone.now)
    created_by = models.ForeignKey(User, related_name="revisions_created")
    media = models.ManyToManyField("MediaFile", blank=True, related_name="revisions")

    def parse(self):
        self.content_html = settings.WIKI_PARSE(self.page.wiki, self.content)

    def process_media(self):
        pks = MEDIA_RE.findall(self.content)
        media = MediaFile.objects.filter(user=self.created_by, pk__in=pks)
        self.media.clear()
        for mf in media:
            self.media.add(mf)

    def save(self, *args, **kwargs):
        super(Revision, self).save(*args, **kwargs)
        self.process_media()

    class Meta:
        get_latest_by = "created_at"


def uuid_filename(instance, filename):
    ext = filename.split(".")[-1]
    filename = "{0}.{1}".format(uuid.uuid4(), ext)
    return os.path.join("revision-files", filename)


class MediaFile(models.Model):

    user = models.ForeignKey(User, related_name="media_files")
    created = models.DateTimeField(editable=False, default=timezone.now)
    filename = models.CharField(max_length=255)
    file = models.FileField(upload_to=uuid_filename)

    def download_url(self):
        return reverse("wiki_file_download", args=[self.pk, os.path.basename(self.filename)])

    def __unicode__(self):
        return self.filename



from __future__ import absolute_import

from django.conf import settings  # noqa
from django.core.exceptions import ImproperlyConfigured
from django.utils import importlib

from appconf import AppConf


def load_path_attr(path):
    i = path.rfind(".")
    module, attr = path[:i], path[i + 1:]
    try:
        mod = importlib.import_module(module)
    except ImportError as e:
        raise ImproperlyConfigured("Error importing {0}: '{1}'".format(module, e))
    try:
        attr = getattr(mod, attr)
    except AttributeError:
        raise ImproperlyConfigured("Module '{0}' does not define a '{1}'".format(module, attr))
    return attr


class WikiAppConf(AppConf):

    BINDERS = [
        "wiki.binders.DefaultBinder"
    ]
    IP_ADDRESS_META_FIELD = "HTTP_X_FORWARDED_FOR"
    HOOKSET = "wiki.hooks.WikiDefaultHookset"
    PARSE = "wiki.parsers.creole_wikiword_parse"

    def configure_binders(self, value):
        binders = []
        for val in value:
            binders.append(load_path_attr(val)())
        return binders

    def configure_hookset(self, value):
        return load_path_attr(value)()

    def configure_parse(self, value):
        return load_path_attr(value)



from .conf import settings


class WikiDefaultHookset(object):

    def can_create_page(self, wiki, user):
        return False

    def can_edit_page(self, page, user):
        return False

    def can_delete_page(self, page, user):
        return False

    def can_view_page(self, page, user):
        return False


class HookProxy(object):

    def __getattr__(self, attr):
        return getattr(settings.WIKI_HOOKSET, attr)


hookset = HookProxy()
